# -*- ruby -*-

#
# VFS for ruby - open anything with URI
# 
#   Copyright: Tatsuki Sugiura <sugi@nemui.org>
#   License: Ruby's original or GPL2
#   $LastChangedDate: 2003-10-09 21:39:59 +0900 (Thu, 09 Oct 2003) $
#   $LastChangedBy: sugi $
#

require 'vfs/common'

module VFS
  class File
    include Enumerable
    include FileBuiltIn::Constants
    include VFS::Common
    extend  VFS::MethodGenerator
    toplevel_methods_to_instance_methods :join, :scheme, :escape, :unescape
    relay_path_to_file_builtin :basename, :dirname, :extname

    raise_unsupported :chmod, :chardev?, :writable_real?, :umask,
      :atime, :pipe?, :exists?, :symlink, :sticky?, :file?, :stat,
      :utime, :blockdev?, :writable?, :ftype, :foreach,
      :grpowned?, :exist?, :link, :setgid?, :executable_real?,
      :delete, :ctime, :socket?, :readable_real?, :truncate, :unlink,
      :lstat, :owned?, :directory?, :chown, :setuid?, :executable?,
      :expand_path, :mtime, :symlink?, :size, :readable?, :readlink,
      :fnmatch
    
    class << self
      def new(url, *args)
	# fixme; split this to common to share with VFS::Dir
	if self == VFS::File
	  scheme = VFS.scheme(url)
	  begin
	    require "vfs/scheme/#{scheme}/file"
	  rescue LoadError
	    raise Error::Unsupported,
	      "scheme '#{scheme}' is not supported yet!"
	  end
	  handler = "::VFS::SchemeHandler::#{scheme.upcase}::File"
	  obj = eval(handler).send :new, url
	  unless obj
	    raise Error::Unsupported,
	      "scheme '#{scheme}' is not supported yet!"
	  end
	  return obj
	else
	  super
	end # self == VFS::File
      end # new

      def utime(n_atime, n_mtime, *files)
	files.each { |f|
	  VFS::File.new(f).utime(n_atime, n_mtime)
	}
      end
    
    end # calss << self
    
    # return new instance. This do not open file.
    def initialize(url)
      @path = url
      @buf  = nil
    end
    attr_reader :path
    alias :url :path

    def size?
      exists? or return nil
      begin
	return nil if size == 0
      rescue Error::Unsupported
	raise
      rescue # fixme: should not catch all exception.
	return nil
      end
      size
    end
    
    def zero?
      begin
	rerun true if size == 0
      rescue Error::Unsupported
	raise
      rescue # fixme: should not catch all exception.
        # ignore simply
      end
      false
    end

    def buf
      require 'vfs/tempbuf'
      @buf == nil ? @buf = VFS::Tempbuf.new(path) : @buf
    end
    private :buf

    def each(rs = $/, &block)
      open(RDONLY) {|io|
	io.each(rs) { |i|
	  yield(i)
	}
      }
    end

    def conv_openflang_to_const(v)
      flag = RDONLY
#      v =~ /b/ and flag |= BINARY # fixme? not exist????
      v =~ /r/ and flag |= RDONLY
      v =~ /w/ and flag |= WRONLY|CREAT|TRUNC
      v =~ /a/ and flag |= WRONLY|CREAT|APPEND
      if v =~ /\+/
	flag &= 65535 & ~(RDONLY|WRONLY)
	flag |= RDWR
      end
      flag
    end
    private :conv_openflang_to_const

    # simple open through File.open. must be override scheme handler
    def open(*args, &block)
      flag = conv_openflang_to_const(args[0])
      if flag & (WRONLY|RDWR) != 0
	open_for_write(*args, &block)
      else
	open_for_read(*args, &block)
      end
    end

    def open_for_read(flag = RDONLY, perm = nil, &block)
      #buf.closed? and buf.open
      #buf.eof? and buf.rewind
      if block_given?
	begin
	  yield(buf)
	ensure
	  buf.close
	end
      else
	buf
      end
    end
    private :open_for_read

    def open_for_write(flag = WRONLY, perm = nil, &block)
      # no default method. must be override by each scheme handler
      raise Error::Unsupported,
	"write open is not supported yet for '#{scheme}' scheme"
    end
    private :open_for_write

    def <<(*args)
      open("a") { |f|
	f << args.to_s
      }
      self
    end

    def rename(dest)
      # FIMXE! rename direcotry is not supported yet.
      if scheme == VFS.scheme(dest)
	rename_internal(dest)
      else
	rename_external(dest)
      end
    end

    def rename_internal(dest)
      log("Lazy rename() in same scheme. Please improvement.", Log::WARNING)
      rename_external(dest)
    end
    private :rename_internal

    def rename_external(dest)
      newfile = VFS::File.new(dest)
      newfile.open("w") { |new|
	each { |l|
	  new << l
	}
      }
      newfile.utime(atime||Time.now, mtime||Time.now)
      unlink
      newfile
    end
    private :rename_external

    instance_methods_to_class_methods(
      *self.instance_methods(false).delete_if{|m| m == "utime" })
  end # VFS::File
end # VFS

