# -*- ruby -*-

#
# VFS for ruby - open anything with URI
# 
#   Copyright: Tatsuki Sugiura <sugi@nemui.org>
#   License: Ruby's original or GPL2
#   $LastChangedDate: 2003-10-09 21:39:59 +0900 (Thu, 09 Oct 2003) $
#   $LastChangedBy: sugi $
#

require 'net/http'

module Net
  class HTTP < Protocol
    def options( path, initheader = nil ) 
      res = request(Options.new(path, initheader))
      res.value unless @newimpl
      res
    end
    def propfind( path, body = nil, initheader = { 'Depth' => '0' } )
      res = request(Propfind.new(path, initheader), body)
      res.value unless @newimpl
      res
    end
    def delete( path, initheader = { 'Depth' => 'infinite' } )
      res = request(Delete.new(path, initheader))
      res.value unless @newimpl
      res
    end
    def move( path, initheader = nil )
      res = request(Move.new(path, initheader))
      res.value unless @newimpl
      res
    end
    def copy( path, initheader = nil )
      res = request(Move.new(path, initheader))
      res.value unless @newimpl
      res
    end
  end # HTTP < Protocol

  class HTTP
    class Options < HTTPRequest
      METHOD = 'OPTIONS'
      REQUEST_HAS_BODY = false
      RESPONSE_HAS_BODY = false
    end
    class Propfind < HTTPRequest
      METHOD = 'PROPFIND'
      REQUEST_HAS_BODY = true
      RESPONSE_HAS_BODY = true
    end
    class Delete < HTTPRequest
      METHOD = 'DELETE'
      REQUEST_HAS_BODY = false
      RESPONSE_HAS_BODY = true
    end
    class Move < HTTPRequest
      METHOD = 'MOVE'
      REQUEST_HAS_BODY = false
      RESPONSE_HAS_BODY = true
    end
    class Copy < HTTPRequest
      METHOD = 'MOVE'
      REQUEST_HAS_BODY = false
      RESPONSE_HAS_BODY = true
    end
  end # HTTP

  class HTTPGenericRequest
    SEND_BUFFER_SIZE = 10240
    def send_request_with_body( sock, ver, path, body )
      begin
	@header['content-length'] = body.length.to_s
      rescue NoMethodError
	begin
	  @header['content-length'] = body.size.to_s
	rescue NoMethodError
	  # give up
	end
      end
      @header.delete 'transfer-encoding'

      unless @header['content-type']
        $stderr.puts 'net/http: warning: Content-Type did not set; using application/x-www-form-urlencoded' if $VERBOSE
        @header['content-type'] = 'application/x-www-form-urlencoded'
      end
p @header
      request sock, ver, path
      if body.kind_of?(IO)
	while ( ! body.eof? )
	  sock.write body.read(SEND_BUFFER_SIZE)
	end
      else
	sock.write body
      end
    end
  end

end

module VFS;  module SchemeHandler;  module HTTP;
  module Common

    PROP_CACHE_TIME = 60

    def check_uri_syntax(url)
      unless parse_host(url)
	raise Error::InvalidURI,
	  "Invalid HTTP URL '#{url}' was speficied."
      end
    end

    def init_common_attr
      @capability   = []
      @prop_cache   = nil
      @prop_cache_t = 0
      @header       = nil
    end
    private :init_common_attr
    attr_writer :capability, :header
    attr_accessor :prop_cache, :prop_cache_t
    private :prop_cache, :prop_cache=, :prop_cache_t, :prop_cache_t=,
      :header=

    def capability?(method)
      capability.member? method      
    end

    def capability
      return @capability unless @capability.length == 0
      capability = http_request(url, :options) { |rep|
	rep["allow"].split(/\s*,\s*/)
      }
    end

    def property(elem)
      unless capability? "PROPFIND"
	return nil
      end
      if prop_cache == nil || Time.now.to_i - prop_cache_t > PROP_CACHE_TIME
	log("Start fetching property.")
	http_request(url, :propfind, nil, {"Depth" => "0"}) { |rep|
	  log("*BUG/NotInplementedYet* must unescape XML entity.", Log::WARNING)
	  __send__ :prop_cache=, rep.body
	}
	__send__ :prop_cache_t=, Time.now.to_i
	log("Property was fetched")
      end
      log("Scaning property '#{elem}'.", Log::DEBUG)
      m = Regexp.new("<#{elem}(?:\s+[^>]*)?>(.*?)</#{elem}>").match(prop_cache)
      if m && m[1]
	m[1].tr("\n\r\t", " ")
      else
	nil
      end
    end

    def header(name=nil)
      if @header == nil
	log("Start fetching header.")
	http_request(url, :head) { |rep|
	  __send__ :header=, rep.each_header{}
	}
	log("Header was fetched")
      end
      name == nil ? @header : @header[name]
    end

    def http_request(target_url, method=:get, *rest_args, &block)
      t_host = parse_host(target_url)
      t_port = parse_port(target_url)
      t_path = VFS.unescape(parse_path(target_url))
      t_auth = VFS.unescape(parse_auth(target_url)) # not supported yet!
      
      Net::HTTP.version_1_1
      begin
	log("Opening HTTP connection.", Log::INFO)
	Net::HTTP::Proxy(parse_host(ENV["parse_proxy"]),
			 parse_port(ENV["parse_proxy"])).start(t_host,
							       t_port) { |http|
	  log("Requesting '#{method}' for '#{t_path}' to server.", Log::DEBUG)
	  response = http.send(method, t_path, *rest_args)
	  Array === response and response = response.first
	  log("Data was fetched from web server. Go yield.", Log::DEBUG)
	  if block_given?
	    yield(response)
	  else
	    response
	  end
	}
      rescue Net::ProtoRetriableError => err
	log("Retriable error was detected in HTTP connection. (detail: #{err})",
	    Log::INFO)
	l = err.response['location']
	if parse_host(l)
	  t_host = parse_host(l)
	  t_port = parse_port(l)
	  t_path = VFS.unescape(parse_path(l))
	  t_auth = VFS.unescape(parse_auth(l)) # always empty?
	  # fxime; never returns when redirect loop.
	  retry
	end
      end
    end # http_request

    private
    # fixme; TODO change parse_* to URI.parse ?
    def parse_path(url)
      Regexp.new("^http://[^/]+(.+)$").match(url).to_a[1]
    end
    
    def parse_host(url)
      Regexp.new("^http://(?:[^/@]+?@)?([^/]+)").match(url).to_a[1]
    end
    
    def parse_port(url)
      Regexp.new("^http://[^/]+:([0-9]+)/").match(url).to_a[1] || 80
    end
    
    def parse_auth(url)
      Regexp.new("^http://([^/]+?)@").match(url).to_a[1]
    end

  end # Common
end; end; end
